
/*--------------------------------------------------------------------
  File:        graphics.cpp
  Author:      Nils Schmeisser
  ----------------------------------------------------------------------
  Description: Compatibility module to Borland's GRAPHICS module
  ----------------------------------------------------------------------
  Notes:       replace #include <graphics.h> by #include "graph.h"
               define USE_GL or USE_X depend on what you want to use,
               either GL or X. If USE_X is defined, set the GRAPHICS_DISPLAY
               to the display variable where you want to put your graphics
               window.
  Date:        22.11.1993
  ----------------------------------------------------------------------
*/ 

#define XTFUNCPROTO
#define ANSI

/*--- For recompiling set the USE_X or USE_GL ----------------*/

/* #define USE_GL */
#define USE_X

/*------------------------------------------------------------*/

#ifdef __cplusplus
extern "C" {
#endif
  #include <stdio.h>
  #include <string.h>
  #include <stdlib.h>
#ifdef __cplusplus
}
#endif
#include <math.h>

#define GRAPH_MODULE
  #include "graphics.h"
#undef GRAPH_MODULE

#ifdef USE_GL
  #include <gl/gl.h>
  #include <gl/device.h>
  
  static long winid;
#endif

#ifdef USE_X
  #include <X11/Xlib.h>
  #include <X11/Xutil.h>
  #include <X11/Xos.h>
  #include <X11/X.h>

  static unsigned long ColorOfs=0;
  static Colormap ColorTable;
  static XColor   ColorEntry[256];

  static Visual *VisualInfo;

  static XEvent XAppGraphicsEvent;

  static Window GRAPHICS_WINDOW;
  static Display *GRAPHICS_DISPLAY=NULL;
  static GC XAppGraphicsContext;
#endif

static byte **imgbuffer=NULL;
static int winon=1;           /* set to 1 if window is opened                                */
static int bbuf=0;            /* set to 1 if backbuffer drawing only, use flush() to display */
int MAXCOLORS;
static XColor dpycolors[256];
int BLACK;
int BLUE;
int GREEN;
int CYAN;
int RED;
int MAGENTA;
int BROWN;
int LIGHTGRAY;
int DARKGRAY;
int LIGHTBLUE;
int LIGHTGREEN;
int LIGHTCYAN;
int LIGHTRED;
int LIGHTMAGENTA;
int YELLOW;
int WHITE;
static unsigned int sizex,sizey;
static int COLOR;
static int penx,peny;

/*--------------------------------------------------------------------------*/

#ifdef USE_X

GC GetGraphicsContext() { return XAppGraphicsContext; } 
Display *GetGraphicsDisplay() { return GRAPHICS_DISPLAY; }
Window GetGraphicsWindow() { return GRAPHICS_WINDOW; }

static int cursor=XC_left_ptr;

void setcursor(int c) {
  Cursor crs=XCreateFontCursor(GRAPHICS_DISPLAY,c);
  XDefineCursor(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,crs);
  cursor=c;
}

void showcursor() {
  setcursor(cursor);
}

void hidecursor() {
  Pixmap pm=XCreatePixmap(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,1,1,1);
  Cursor crs=XCreatePixmapCursor(GRAPHICS_DISPLAY,pm,None,
				 &ColorEntry[WHITE],&ColorEntry[BLACK],0,0);
  XDefineCursor(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,crs);
}

void ClearEvent(XEvent *event) {
  if (event!=NULL) 
    XNextEvent(GRAPHICS_DISPLAY,event);
}

void CheckEvent(XGraphicsEventHandler handler,void *data) {
  XEvent event;
  int x,y;
  unsigned int border,depth;
  Window rootw;
  char buffer[256];
  KeySym key;
  XComposeStatus compose;

  if (XCheckWindowEvent(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,
			ButtonPressMask | ButtonReleaseMask | PointerMotionMask | KeyPressMask |
			ColormapChangeMask | ExposureMask | ResizeRedirectMask | LeaveWindowMask |
			EnterWindowMask,
			&event)) {

    if (event.type==ButtonPress) {
    } else if (event.type==KeyPress) {
      XLookupString(&(event.xkey),buffer,256,&key,&compose);
      event.xkey.keycode=key;
    } else if (event.type==LeaveNotify) {
    } else if (event.type==EnterNotify) {
    } else {
      XGetGeometry(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,&rootw,&x,&y,&sizex,&sizey,&border,&depth);
    }
    handler(&event,data);
  }
}

inline unsigned long myabs(int i) {
  if (i<0)
    return -i;
  else
    return i;
}

int ConvRGBtoPIX(int red,int green,int blue) {
  XColor q;
  int i;
  double dr,dg,db;
  double dist,min=256.0*256.0*3.0+1.0;
  int c;

  c=0;
  for (i=0;i<MAXCOLORS+ColorOfs;i++) {
    dr=(double)(dpycolors[i].red-red);
    dg=(double)(dpycolors[i].green-green);
    db=(double)(dpycolors[i].blue-blue);
    dist=dr*dr+dg*dg+db*db;
    if (dist<min) {
      min=dist;
      c=i;
    }
  }
  return(c-ColorOfs);
}
#endif

void setrgbpalette(int colornum, int red, int green, int blue) {
  ColorEntry[colornum+ColorOfs].pixel  =(unsigned long)(colornum+ColorOfs);
  ColorEntry[colornum+ColorOfs].red    =(unsigned short)(red*256);
  ColorEntry[colornum+ColorOfs].green  =(unsigned short)(green*256);
  ColorEntry[colornum+ColorOfs].blue   =(unsigned short)(blue*256);
  ColorEntry[colornum+ColorOfs].flags  = DoRed | DoGreen | DoBlue;    
  if ((winon) && (bbuf==0))
  #ifdef USE_GL
    mapcolor(colornum,red,green,blue);
  #endif
  #ifdef USE_X  
    XStoreColor(GRAPHICS_DISPLAY,ColorTable,&ColorEntry[colornum+ColorOfs]);
  #endif
}

void SetColors_to_VGA() {
  setrgbpalette(BLACK       ,  0,  0,  0);
  setrgbpalette(BLUE        ,  0,  0,127);
  setrgbpalette(GREEN       ,  0,127,  0);
  setrgbpalette(CYAN        ,  0,127,127);
  setrgbpalette(RED         ,127,  0,  0);
  setrgbpalette(MAGENTA     ,127,  0,127);
  setrgbpalette(BROWN       ,127,127,  0);
  setrgbpalette(LIGHTGRAY   ,127,127,127);
  setrgbpalette(DARKGRAY    , 63, 63, 63);
  setrgbpalette(LIGHTBLUE   ,  0,  0,255);
  setrgbpalette(LIGHTGREEN  ,  0,255,  0);
  setrgbpalette(LIGHTCYAN   ,  0,255,255);
  setrgbpalette(LIGHTRED    ,255,  0,  0);
  setrgbpalette(LIGHTMAGENTA,255,  0,255);
  setrgbpalette(YELLOW      ,255,255,  0);
  setrgbpalette(WHITE       ,255,255,255);
}

void cleardevice() {
  if ((winon) && (bbuf==0)) {
  #ifdef USE_GL
    color(BLACK);
    clear();
  #endif
  #ifdef USE_X
    XClearWindow(GRAPHICS_DISPLAY,GRAPHICS_WINDOW);
  #endif
  }
  unsigned int i;
  if (imgbuffer!=NULL)
    for (i=0;i<sizey;i++)
      memset((void*)(imgbuffer[i]),0,sizex);
}

#ifdef USE_X
int XGraphicsOn(char *dpy) {
  if (dpy==NULL)
    GRAPHICS_DISPLAY=XOpenDisplay((char*)NULL);
  else
    GRAPHICS_DISPLAY=XOpenDisplay(dpy);
  if (GRAPHICS_DISPLAY!=NULL)
    return 1;
  else
    return 0;
}

void XGraphicsOff() {
  XCloseDisplay(GRAPHICS_DISPLAY);
}
#endif

/*--------------------------------------------------------------------------*/

void backbuffer(int onoff) {
  bbuf=onoff;
}

/*--------------------------------------------------------------------------*/

#ifdef USE_X
  Window initgraph(int *d,int *m,char *path,byte **buffer,int dpymode) {
#else
  long int initgraph(int *d,int *m,char *path,byte **buffer,int dpymode) {
#endif
  char driver[40];
  #ifdef USE_X
    int i;
    Window root;
    int depth,screen;
    unsigned int border_width = 0;
    int x=0,y=0; 
    XColor q;

    XGCValues GCValues;
    XSetWindowAttributes Attr;
  #endif

  imgbuffer=buffer;
  switch (*d) {
   case VGA: switch (*m) {
               case VGAFREE:  sscanf(path,"%d*%d*%d",&sizex,&sizey,&MAXCOLORS);
                              sprintf(driver,"Free VGA %s",path);
			      break;
               case VGAHI:    sizex=640;
                              sizey=480;
                              MAXCOLORS=16;
                              strcpy(driver,"Standard VGA 640*480*16");
                              break;
               case VGAHI256: sizex=640;
                              sizey=480;
                              MAXCOLORS=256;
                              strcpy(driver,"Super VGA 640*480*256");
                              break;
               case   SVGA:   sizex=800;
                              sizey=600;
                              MAXCOLORS=16;
                              strcpy(driver,"Super VGA 800*600*16");
                              break;
               case SVGA256:  sizex=800;
                              sizey=600;
                              MAXCOLORS=256;
                              strcpy(driver,"Super VGA 800*600*256");
                              break;
               case SVGATRUE: sizex=800;
                              sizey=600;
                              MAXCOLORS=16200;   /* Means 16.2 Million */
                              strcpy(driver,"Super VGA 800*600*16.2 Mill");
                              break;
	     }
             break;

  }

  if ((buffer==NULL) || ((buffer!=NULL) && (dpymode==1))) {
  #ifdef USE_GL
    prefposition(10,long(sizex)+10,1000-long(sizey),1000);
    winid=winopen(driver);
    if ((*m)==SVGATRUE) {
      RGBmode();
      gconfig();
    }
    SetColors_to_VGA();
  #endif
  #ifdef USE_X
    if (GRAPHICS_DISPLAY==NULL)
      XGraphicsOn();
    screen=XDefaultScreen(GRAPHICS_DISPLAY);
    root  =XDefaultRootWindow(GRAPHICS_DISPLAY);
    depth =XDefaultDepth(GRAPHICS_DISPLAY,screen);
    VisualInfo=XDefaultVisual(GRAPHICS_DISPLAY,screen);

    ColorTable=XDefaultColormap(GRAPHICS_DISPLAY,screen);

    for (i=0;i<MAXCOLORS;i++) {
      q.pixel=(unsigned long)i;
      XQueryColor(GRAPHICS_DISPLAY,ColorTable,&q);
      ColorEntry[i]=q;
      dpycolors[i].red  =q.red  /256;
      dpycolors[i].green=q.green/256;
      dpycolors[i].blue =q.blue /256;
    }

    BLACK       =ConvRGBtoPIX(  0,  0,  0);
    BLUE        =ConvRGBtoPIX(  0,  0,127);
    GREEN       =ConvRGBtoPIX(  0,127,  0);
    CYAN        =ConvRGBtoPIX(  0,127,127);
    RED         =ConvRGBtoPIX(127,  0,  0);
    MAGENTA     =ConvRGBtoPIX(127,  0,127);
    BROWN       =ConvRGBtoPIX(127,127,  0);
    LIGHTGRAY   =ConvRGBtoPIX(127,127,127);
    DARKGRAY    =ConvRGBtoPIX( 63, 63, 63);
    LIGHTBLUE   =ConvRGBtoPIX(  0,  0,255);
    LIGHTGREEN  =ConvRGBtoPIX(  0,255,  0);
    LIGHTCYAN   =ConvRGBtoPIX(  0,255,255);
    LIGHTRED    =ConvRGBtoPIX(255,  0,  0);
    LIGHTMAGENTA=ConvRGBtoPIX(255,  0,255);
    YELLOW      =ConvRGBtoPIX(255,255,  0);
    WHITE       =ConvRGBtoPIX(255,255,255);

    Attr.event_mask=ResizeRedirectMask | ExposureMask | ButtonPressMask | ColormapChangeMask | 
                    LeaveWindowMask | EnterWindowMask;
    Attr.background_pixel=0;
    Attr.border_pixel    =1;
    Attr.colormap  =ColorTable;
    GRAPHICS_WINDOW=XCreateWindow(GRAPHICS_DISPLAY,root,
                                  x,y,sizex,sizey,border_width,depth,
                                  InputOutput,VisualInfo,
                                  CWBackPixel | CWEventMask | CWBorderPixel | CWColormap,&Attr);
  
    GCValues.function  =GXcopy;
    GCValues.plane_mask=AllPlanes;
    GCValues.foreground=1;
    GCValues.background=0;
    XAppGraphicsContext=XCreateGC(GRAPHICS_DISPLAY, root /* GRAPHICS_WINDOW */ ,
                                  GCFunction | GCPlaneMask | GCForeground | 
				  GCBackground,&GCValues); 
    XMapWindow(GRAPHICS_DISPLAY,GRAPHICS_WINDOW);
    XInstallColormap(GRAPHICS_DISPLAY,ColorTable);
    XSelectInput(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,
                 ButtonPressMask | ButtonReleaseMask | PointerMotionMask | KeyPressMask |
                 ColormapChangeMask | ExposureMask | LeaveWindowMask | EnterWindowMask);
    showcursor();
    XNextEvent(GRAPHICS_DISPLAY,&XAppGraphicsEvent);
  #endif
  } else {
    SetColors_to_VGA();
    #ifdef USE_X
      GRAPHICS_WINDOW=0;
    #else
      winid=0;
    #endif
    winon=0;
  }

  COLOR=WHITE;
  cleardevice();
  #ifdef USE_X
    return GRAPHICS_WINDOW;
  #else
    return winid;
  #endif
}

void moveto(int x,int y) {
  penx=x;
  peny=y;
}

void lineto(int x,int y) {
  line(penx,peny,x,y);
  penx=x;
  peny=y;
}

//-----------------------------------------------------------------------

void arc(int x,int y,int startangle,int stopangle,int r) {
  if ((winon) && (bbuf==0)) {
  #ifdef USE_X
    XSetForeground(GRAPHICS_DISPLAY,XAppGraphicsContext,COLOR+ColorOfs);
    XDrawArc(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,XAppGraphicsContext,
	     x-r,y-r,2*r,2*r,startangle*64,stopangle*64);
//    XFlush(GRAPHICS_DISPLAY);
  #endif
  }
  if (imgbuffer!=NULL) {
  }
}

//-----------------------------------------------------------------------

void circle(int x,int y,int r) { arc(x,y,r,0,360); }

void closegraph() {
  if ((winon) && (bbuf==0)) {
  #ifdef USE_GL
    winclose(winid);
  #endif
  #ifdef USE_X
    XUninstallColormap(GRAPHICS_DISPLAY,ColorTable);
    XFreeColormap(GRAPHICS_DISPLAY,ColorTable);
    XFreeGC(GRAPHICS_DISPLAY,XAppGraphicsContext);
    XDestroyWindow(GRAPHICS_DISPLAY,GRAPHICS_WINDOW);	
    XGraphicsOff();
  #endif
  }
}

int getmaxcolor() {
  return(MAXCOLORS);
}

int getmaxx() {
  return(sizex);
}

int getmaxy() {
  return(sizey);
}

void setcolor(int color) {
  COLOR=color;
}

void putpixel(int x,int y,int col) {
  if ((winon) && (bbuf==0)) {
  #ifdef USE_GL
    color(col);
    rect(x,getmaxy()-y,x+1,getmaxy()-y-1);
  #endif
  #ifdef USE_X
    XSetForeground(GRAPHICS_DISPLAY,XAppGraphicsContext,col+ColorOfs);
    XDrawPoint(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,XAppGraphicsContext,x,y);
  #endif
  }
  if (imgbuffer!=NULL) {
    imgbuffer[y][x]=col+ColorOfs;
  }
}

void line(int x1,int y1,int x2,int y2) {

  if ((winon) && (bbuf==0)) {
  #ifdef USE_GL
    long p1[2],p2[2];

    color(COLOR);
    p1[0]=x1;
    p1[1]=getmaxy()-y1;
    p2[0]=x2;
    p2[1]=getmaxy()-y2;
    bgnline();
      v2i(p1);
      v2i(p2);
    endline();
  #endif
  #ifdef USE_X
    XSetForeground(GRAPHICS_DISPLAY,XAppGraphicsContext,COLOR+ColorOfs);
    XDrawLine(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,XAppGraphicsContext,x1,y1,x2,y2);	
//    XFlush(GRAPHICS_DISPLAY);
  #endif
  }
  if (imgbuffer!=NULL) {
    /* the algorithm of BRESENHAM */

    int x,y,h;
    int d,dx,dy;
    int sgny,swp;
    int incrE,incrNE;
    int col=COLOR+ColorOfs;

    if (y1==y2) {
      dx=x2-x1;
      if (dx>=0)
	memset(imgbuffer[y1]+x1,col,dx+1);
      else
	memset(imgbuffer[y1]+x2,col,-dx+1);
    } else if (x1==x2) {
      dy=y2-y1;
      if (dy>=0)
	for (y=y1;y<=y2;y++)
	  *(*(imgbuffer+y)+x1)=col;
      else
	for (y=y2;y<=y1;y++)
	  *(*(imgbuffer+y)+x1)=col;
    } else {
      if (x2<x1) {
	h=x1;
	x1=x2;
	x2=h;
	h=y1;
	y1=y2;
	y2=h;
      }
      dx=x2-x1;
      dy=y2-y1;
      if (dy<0) {
	sgny=-1;
	dy=-dy;
      } else
	sgny=1;
      if (dy>dx) {
	h=dx;
	dx=dy;
	dy=dx;
	swp=1;
      } else
	swp=0;
      d=dy+dy-dx;
      incrE=dy;
      incrE=incrE+incrE;
      incrNE=dy-dx;
      incrNE=incrNE+incrNE;
      x=x1;
      y=y1;
      *(*(imgbuffer+y)+x)=(byte)col;
      while (x<x2) {
	if (d<=0) {
	  d+=incrE;
	  if (swp)
	    y++;
	  else
	    x++;
	} else {
	  d+=incrNE;
	  if (swp) {
	    y++;
	    x+=sgny;
	  } else {
	    x++;
	    y+=sgny;
	  }
	}
	*(*(imgbuffer+y)+x)=(byte)col;
      }
    }
  }
}

void putRGBpixel(int x,int y,int r,int g,int b) {
  if ((winon) && (bbuf==0)) {
  #ifdef USE_GL
    RGBcolor(r,g,b);
    rect(x,getmaxy()-y,x+1,getmaxy()-y-1);
  #endif
  #ifdef USE_X
    XSetForeground(GRAPHICS_DISPLAY,XAppGraphicsContext,ConvRGBtoPIX(r,g,b));
    XDrawPoint(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,XAppGraphicsContext,x,y);
//    XFlush(GRAPHICS_DISPLAY);
  #endif
  }
}

unsigned long imagesize(int x1,int y1,int x2,int y2) {
  unsigned long s;

  s=(unsigned long)(x2-x1+1)*(unsigned long)(y2-y1+1)+4;
  return s;
}

void outtextxy(int x,int y,char *s,int color) {
  if ((winon) && (bbuf==0)) {
  #ifdef USE_X
    XTextItem text;

    if (color>-1) 
      XSetForeground(GRAPHICS_DISPLAY,XAppGraphicsContext,color+ColorOfs);

    text.chars=s;
    text.nchars=strlen(s);
    text.delta=1;
    text.font=None;
    XDrawText(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,XAppGraphicsContext,x,y,&text,1);
//    XFlush(GRAPHICS_DISPLAY);
  #endif
  }
}

void bar(int x0,int y0,int x1,int y1,int col) {
  if ((winon) && (bbuf==0)) {
  #ifdef USE_X
    if (col>-1) 
      XSetForeground(GRAPHICS_DISPLAY,XAppGraphicsContext,col+ColorOfs);
    XFillRectangle(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,XAppGraphicsContext,x0,y0,x1-x0+1,y1-y0+1);
//    XFlush(GRAPHICS_DISPLAY);
  #endif
  }
}

void rectangle(int x0,int y0,int x1,int y1,int col) {
  if ((winon) && (bbuf==0)) {
  #ifdef USE_X
    if (col>-1) 
      XSetForeground(GRAPHICS_DISPLAY,XAppGraphicsContext,col+ColorOfs);
    XDrawRectangle(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,
		   XAppGraphicsContext,x0,y0,x1-x0+1,y1-y0+1);
//    XFlush(GRAPHICS_DISPLAY);
  #endif
  }
}

/*
void getimage(int x1,int y1,int x2,int y2,void *bitmap) {
  unsigned short w,h,*d;
  char *data;

  data=(char*)bitmap;
  d=(unsigned short*)bitmap;
  data++;
  data++;
  data++;
  data++;

  w=x2-x1+1;
  h=y2-y1+1;
  (*d)=w;
  d++;
  (*d)=h;

  #ifdef USE_GL
    Int16 *dat;
    long i,j;

    dat=new Int16 [w*h];
    rectread(x1,y1,x2,y2,dat);
    for (j=h-1;j>=0;j--) {
      for (i=0;i<w;i++) {
        data[((h-1)-j)*w+i]=(char)dat[j*w+i];
      }
    }
    delete dat;
  #endif
  #ifdef USE_X
    XImage *p;  
    unsigned long i,j,o;

    p=XCreateImage(GRAPHICS_DISPLAY,VisualInfo,8,ZPixmap,0,data,w,h,8,w);
    (*p)=*XGetImage(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,x1,y1,w,h,0xffffffff,ZPixmap);
    o=p->bytes_per_line-w;
    for (i=0;i<h;i++) for (j=0;j<w;j++) data[i*w+j]=p->data[i*(w+o)+j];
  #endif
}

void putimage(int x,int y,void *bitmap,int flag) {
  unsigned short w,h,*d;
  char *data;

  data=(char*)bitmap;
  d=(unsigned short*)bitmap;
  data++;
  data++;
  data++;
  data++;
  w=(*d);
  d++;
  h=(*d);

  #ifdef USE_GL
    Int16 *dat;
    long i,j;

    dat=new Int16 [w*h];
    for (j=h-1;j>=0;j--) {
      for (i=0;i<w;i++) {
        dat[j*w+i]=(Int16)data[((h-1)-j)*w+i];
      }
    }
    rectwrite(x,y,x+w-1,y+h-1,dat);
    delete dat;
  #endif
  #ifdef USE_X
    XImage *p;
    p=XCreateImage(GRAPHICS_DISPLAY,VisualInfo,8,ZPixmap,0,data,w,h,8,w);
    XPutImage(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,XAppGraphicsContext,p,0,0,x,y,w,h);
  #endif
}
*/

byte **createbuffer(int maxx,int maxy) {
  int i;
  byte **buf;

  buf=new byte* [maxy];
  for(i=0;i<maxy;i++)
    buf[i]=new byte [maxx];
  return buf;
}

void setbuffer(byte **buffer) {
  imgbuffer=buffer;
}

void savebuffer(char *fname,int fmode) {
  if (imgbuffer!=NULL) {
    if (fmode==BMP) {
    }
  }
}

void flush() {
  #ifdef USE_X
    XFlush(GRAPHICS_DISPLAY);
  #endif
}
