/*--------------------------------------------------------------------
  File:        graphics.h
  Author:      Nils Schmeisser
  ----------------------------------------------------------------------
  Description: Compatibility module to Borland's GRAPHICS module, include
  ----------------------------------------------------------------------
  Notes:       replace #include <graphics.h> by #include "graph.h"
               define USE_GL or USE_X depend on what you want to use,
               either GL or X, in "deflib.h".
               If USE_X is defined, set the GRAPHICS_DISPLAY to the 
               display variable where you want to put your graphics
               window.
  Date:        22.11.1993
  ----------------------------------------------------------------------
*/ 

#ifndef GRAPHICS_
#define GRAPHICS_

#include <stdlib.h>

#ifndef byte
  typedef unsigned char byte;
#endif

/* Graphics drivers */

#define DETECT   0
#define CGA      1
#define MCGA     2
#define EGA      3
#define EGA64    4
#define EGAMONO  5
#define IBM8514  6
#define HERCMONO 7
#define ATT400   8
#define VGA      9
#define PC3270   10

/* Graphics modes */

#define CGAC0     0 /* 320*200 palette 0 */
#define CGAC1     1 /* 320*200 palette 1 */
#define CGAC2     2 /* 320*200 palette 2 */
#define CGAC3     3 /* 320*200 palette 3 */
#define CGAHI     4 /* 640*200*2 */

#define MCGAC0    0 /* 320*200 palette 0 */
#define MCGAC1    1 /* 320*200 palette 1 */
#define MCGAC2    2 /* 320*200 palette 2 */
#define MCGAC3    3 /* 320*200 palette 3 */
#define MCGAMED   4 /* 640*200*2 */
#define MCGAHI    5 /* 640*480*2 */

#define VGAFREE   1 /* free definable */
#define VGAHI     2 /* 640*480*16 */
#define VGAHI256  3 /* 640*480*256 */
#define SVGA      4 /* 800*600*16 */
#define SVGA256   5 /* 800*600*256 */
#define SVGATRUE  6 /* 800*600*16.2Mill */

extern int BLACK;
extern int BLUE;
extern int GREEN;
extern int CYAN;
extern int RED;
extern int MAGENTA;
extern int BROWN;
extern int LIGHTGRAY;
extern int DARKGRAY;
extern int LIGHTBLUE;
extern int LIGHTGREEN;
extern int LIGHTCYAN;
extern int LIGHTRED;
extern int LIGHTMAGENTA;
extern int YELLOW;
extern int WHITE;

  #ifdef USE_X
    #include <X11/Xlib.h>
    #include <X11/Xutil.h>
    #include <X11/Xos.h>
    #include <X11/X.h>
    #include <X11/keysym.h>
    #include <X11/cursorfont.h>
    
//    extern GC XAppGraphicsContext;
//    extern Display *GRAPHICS_DISPLAY;
//    extern Window GRAPHICS_WINDOW;

    typedef void (*XGraphicsEventHandler)(XEvent *event,void *data);
    extern void ClearEvent(XEvent *event);
    extern void CheckEvent(XGraphicsEventHandler handler,void *data);
    extern int XGraphicsOn(char *dpy=NULL);
    extern void XGraphicsOff();
    
    extern GC GetGraphicsContext();
    extern Display *GetGraphicsDisplay();
    extern Window GetGraphicsWindow();
  #endif

  extern int MAXCOLORS;

  #ifdef USE_X
    extern Window initgraph(int *d,int *m,char *path,byte **buffer=NULL,int dpymode=0);
    extern void setcursor(int);
    extern void showcursor();
    extern void hidecursor();
  #else
    extern long int initgraph(int *d,int *m,char *path,byte **buffer=NULL,int dpymode=0);
  #endif

  extern void backbuffer(int onoff=1);

  extern void moveto(int x,int y);
  extern void lineto(int x,int y);
  extern void circle(int x,int y,int r);
  extern void closegraph();
  extern void cleardevice();

  extern int getmaxx();
  extern int getmaxy();

  extern void setrgbpalette(int colornum, int red, int green, int blue);
  extern void setcolor(int color);
  extern int getmaxcolor();

  extern void putpixel(int x,int y,int col);
  extern void line(int x1,int y1,int x2,int y2);
  extern void putRGBpixel(int x,int y,int r,int g,int b);

  extern unsigned long imagesize(int x1,int y1,int x2,int y2);
  extern void getimage(int x1,int y1,int x2,int y2,void *bitmap);
  extern void putimage(int x,int y,void *bitmap,int flag);

  extern void outtextxy(int x,int y,char *s,int color=-1);
  extern void bar(int x0,int y0,int x1,int y1,int col=-1);
  extern void rectangle(int x0,int y0,int x1,int y1,int col=-1);

  #define BMP 0

  extern void setbuffer(byte **buffer);
  extern void savebuffer(char *fname,int fmode=BMP);
  extern void flush();
#endif
