#ifndef LCDAPI_
#define LCDAPI_

  #ifndef byte
    typedef unsigned char byte;
  #endif
  #ifndef word
    typedef unsigned short word;
  #endif

  #define LEFT  0
  #define RIGHT 1

  #ifdef LCDAPIMODULE

  #ifndef USE_X
    #define USE_X
  #endif

  #include <graphics.h>

  typedef enum {GE_POINT,GE_LINE,GE_RECT,GE_POLY,GE_CIRCLE,GE_ARC,GE_ELLIPSE,GE_ARCELLIPSE,
		GE_FILLED_RECT,GE_FILLED_POLY,GE_FILLED_CIRCLE,GE_FILLED_ARC,
		GE_FILLED_ELLIPSE,GE_FILLED_ARCELLIPSE} GEtype;

  class GE {
  public:
    GEtype type;
    virtual void Draw() {};
  };

  class GEpoint:public GE {
  public:
    int x,y,c;
    inline GEpoint(int x_,int y_,int c_) { x=x_;y=y_;c=c_; };
    inline void Draw() { putpixel(x,y,c); };
  };

  class GEline:public GE {
  public:
    int x1,y1,x2,y2,c;
    inline GEline(int xa,int ya,int xb,int yb,int c_) {
      x1=xa;y1=ya;x2=xb;y2=yb;c=c_;
    };
    inline void Draw() { setcolor(c); line(x1,y1,x2,y2); };
  };

  class GErect:public GE {
  public:
    int x1,y1,x2,y2,c;
    inline GErect(int xa,int ya,int xb,int yb,int c_) {
      x1=xa;y1=ya;x2=xb;y2=yb;c=c_;
    };
    inline void Draw() { setcolor(c); rectangle(x1,y1,x2,y2); };
  };
  typedef GErect GEfilledrect;

  class GEpoly:public GE {
  public:
    unsigned int n,c;
    int *x,*y;
    inline GEpoly(int n_,int *x_,int *y_) {
      n=n_;x=x_;y=y_;
    };
    inline void Draw() {};
  };
  typedef GEpoly GEfilledpoly;

  class GEcircle:public GE {
  public:
    int xm,ym,r,c;
    inline GEcircle(int xm_,int ym_,int r_,int c_) { xm=xm_;ym=ym_;r=r_;c=c_;};
    inline void Draw() { setcolor(c); circle(xm,ym,r); };
  };
  typedef GEcircle GEfilledcircle;

  class GEarc:public GE {
  public:
    int xm,ym,r,w1,w2,c;
    inline GEarc(int xm_,int ym_,int r_,int w1_,int w2_,int c_) {
      xm=xm_;ym=ym_;r=r_;w1=w1_;w2=w2_;c=c_;
    };
    inline void Draw() {};
  };
  typedef GEarc GEfilledarc;

  class GEellipse:public GE {
  public:
    int xm,ym,a,b,c;
    inline GEellipse(int xm_,int ym_,int a_,int b_,int c_) { 
      xm=xm_;ym=ym_;a=a_;b=b_;c=c_; 
    };
    inline void Draw() {};
  };
  typedef GEellipse GEfilledellipse;

  class GEarcellipse:public GE {
  public:
    int xm,ym,a,b,w1,w2,c;
    inline GEarcellipse(int xm_,int ym_,int a_,int b_,int w1_,int w2_,int c_) {
      xm=xm_;ym=ym_;a=a_;b=b_;w1=w1_;w2=w2_;c=c_;
    };
    inline void Draw() {};
  };
  typedef GEarcellipse GEfilledarcellipse;

  extern void delay(unsigned long msec);

  extern void Initialize(word sizex,word sizey,word colordepth,
			 word Tlcd=0,word Tcrt=0,word Tv=0);
  extern void Reset();
  extern void ClearCRT();
  extern void SetTimings(float Tlcd,float Tcrt,float Tv);
  extern void GetTimings(float *Tlcd,float *Tcrt,float *Tv);

  extern void SetBuffer(int pg);
  extern void FlushBuffer();

  extern void SetPage(int pg);
  extern void ClearPage(int pg=-1);
  extern void FlushPage();

  extern void SetColor(int c);
  extern void SetFillstyle(int st,int c);

  extern void DrawPixel(int pg,int x,int y,int c=-1);
  extern void DrawLine(int pg,int x1,int y1,int x2,int y2,int c=-1);
  extern void DrawRectangle(int pg,int x1,int y1,int x2,int y2,int c=-1);
  extern void DrawPolygon(unsigned int n,int *x,int *y,int c=-1);
  extern void DrawCircle(int pg,int xm,int ym,int r,int c=-1);
  extern void DrawArc(int pg,int xm,int ym,int r,int w1,int w2,int c=-1);
  extern void DrawEllipse(int pg,int xm,int ym,int a,int b,int c=-1);
  extern void DrawArcEllipse(int pg,int xm,int ym,int a,int b,int w1,int w2,int c=-1);

  extern void FillRectangle(int pg,int x1,int y1,int x2,int y2,int c=-1);
  extern void FillPolygon(unsigned int n,int *x,int *y,int c=-1);
  extern void FillCircle(int pg,int xm,int ym,int r,int c=-1);
  extern void FillArc(int pg,int xm,int ym,int r,int w1,int w2,int c=-1);
  extern void FillEllipse(int pg,int xm,int ym,int a,int b,int c=-1);
  extern void FillArcEllipse(int pg,int xm,int ym,int a,int b,int w1,int w2,int c=-1);

  #endif

  typedef enum {
    LCDAPI_RESET,LCDAPI_SETTIMINGS,LCDAPI_GETTIMINGS,
    LCDAPI_SETPAGE,LCDAPI_CLRPAGE,LCDAPI_CLRCRT,LCDAPI_FLUSH,

    LCDAPI_SETCOLOR,LCDAPI_SETFILLSTYLE,

    LCDAPI_DRAW_PIXEL,LCDAPI_DRAW_LINE,LCDAPI_DRAW_RECT,LCDAPI_DRAW_POLY,
    LCDAPI_DRAW_CIRCLE,LCDAOI_DRAW_ARC,LCDAPI_DRAW_ELLIPSE,LCDAPI_DRAW_ARCELLIPSE,

    LCDAPI_FILL_RECT,LCDAPI_FILL_POLY,LCDAPI_FILL_CIRCLE,LCDAPI_FILL_ARC,
    LCDAPI_FILL_ELLIPSE,LCDAPI_FILL_ARCELLIPSE
  } LCDAPICMD;

  
#endif
