#define USE_X
#define LCDAPIMODULE
#include <lcdapi.h>
#include <graphics.h>
#include <stdio.h>

#ifdef __cplusplus
  extern "C" {
#endif
#include <lcdshutter.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#ifdef __cplusplus
  }
#endif

#include <string.h>

#ifdef USE_X
  static Display *dpy;
  static Window win;
  static GC gc;
#endif

static GE **lbuffer=NULL,**rbuffer=NULL,**buffer=NULL;
static GE **lpage=NULL,**rpage=NULL;
static GE **page=NULL;
static unsigned int nelements,lelements=0,relements=0,lpn=0,rpn=0;
static unsigned int MAXELEMENTS=4096;

static word maxx=0,maxy=0,maxc=0;
static word tlcd=0,tcrt=0,tv=0;
static int channel;

//static pid_t lcdd;

/*--- the LCD switch part ---*/

static unsigned long DELAYCYC;    /*--- cycles per msec ---*/

void delay(unsigned long msec) {
  unsigned long i,n;

  n=DELAYCYC*msec;
  for (i=0;i<n;i++);
}

static int side=LEFT;

void Initialize(word sizex,word sizey,word colordepth,
		word Tlcd,word Tcrt,word Tv) {
  unsigned long n;
  clock_t c1,c2,dc;
  char s[256];
  int d,m; //,result;
  FILE *f;
  key_t msg_key;
  int msg_flag,i;

  /*--- calibrating delay loop ---*/
  printf("Calibrating delay loop\n");
  DELAYCYC=1;
  dc=0;
  while (dc<10) {
    DELAYCYC*=10;
    c1=clock();
    delay(1);
    c2=clock();
    dc=c2-c1;
  }
  n=DELAYCYC;
  DELAYCYC=(n*CLOCKS_PER_SEC)/dc/1000;
  printf("delay loop: DELAYCYC=%d\n",(unsigned int)DELAYCYC);
  printf("initializing communication\n");
  sprintf(s,"/tmp/lcdd.%d",getuid());
  f=fopen(s,"r");
  if (f!=NULL) {
    fclose(f);
    printf("an error occured while trying to create %s\nmaybe lcdd is almost loaded\n",s);
    exit(-1);
  }
  maxx=sizex;
  maxy=sizey;
  maxc=colordepth;
  printf("initializing graphics subsystem\n");
  i=0;
  do {
    i++;
    msg_key=(key_t)i;
    msg_flag=IPC_CREAT | 0666;
    channel=msgget(msg_key,msg_flag);
  } while ((channel==-1) && (i<256));
  if ((i>=256) && (channel==-1)) {
    printf("error opening communication channel, terminating\n");
    exit(-1);
  }
  printf("using message (%d,%d)\n",channel,i);
  f=fopen(s,"w");
  fprintf(f,"%d:%d",getpid(),i);
  fclose(f);
  d=VGA;
  m=VGAFREE;
  sprintf(s,"%d*%d*%d",maxx,maxy,maxc);
  #ifdef USE_X
    XGraphicsOn();
  #endif
  initgraph(&d,&m,s);
  #ifdef USE_X
    dpy=GetGraphicsDisplay();
    win=GetGraphicsWindow();
    gc=GetGraphicsContext();
  #endif
  setcolor(WHITE);
  printf("allocating buffer for %d elements = 2 * %d kByte\n",
	 MAXELEMENTS,2*sizeof(GE*)*MAXELEMENTS/1024);
  lbuffer=new GE* [MAXELEMENTS];
  rbuffer=new GE* [MAXELEMENTS];
  printf("setting timings\n");
  if (Tv>0) {
    tlcd=Tlcd;
    tcrt=Tcrt;
    tv=Tv;
    printf("maximum refreshing rate %4.1f Hz\n",500.0/(tlcd+tcrt+tv));
  }
  printf("clearing CRT and buffers\n");
  ClearCRT();
  SetPage(RIGHT);
  ClearPage();
  SetPage(LEFT);
  ClearPage();
  ShutterInit();
}

void ClearCRT() {
  cleardevice();
  flush();
}

void SetPage(int pg) {
  if (pg==LEFT) {
    page=lpage;
    nelements=lpn;
  } else {
    page=rpage;
    nelements=rpn;
  }
}


void FlushBuffer() {
  
  if (lpage!=NULL)
    delete lpage;
  lpage=lbuffer;
  lpn=lelements;
  lelements=0;
  lbuffer=new GE* [MAXELEMENTS];

  if (rpage!=NULL)
    delete rpage;
  rpage=rbuffer;
  rpn=relements;
  relements=0;
  rbuffer=new GE* [MAXELEMENTS];

  if (page==lpage) {
    page=lpage;
    nelements=lpn;
  } else {
    page=rpage;
    nelements=rpn;
  }
}

void FlushPage() {
  unsigned long i;
  
  for (i=0;i<nelements;i++) {
    page[i]->Draw();
  }
  flush();
}

void ClearPage(int pg) {
  unsigned int i;

  if (pg==-1) {
    for (i=0;i<nelements;i++)
      delete buffer[i];
    if (buffer==lbuffer)
      lelements=0;
    else
      relements=0;
    nelements=0;
  } else if (pg==LEFT) {
    for (i=0;i<lelements;i++)
      delete lbuffer[i];
    lelements=0;
    if (buffer==lbuffer)
      nelements=0;
  } else {
    for (i=0;i<relements;i++)
      delete rbuffer[i];
    relements=0;
    if (buffer==rbuffer)
      nelements=0;
  }
  FlushBuffer();
}

void AddGE(int pg,GE *item) {
  if (pg==LEFT) {
    if (lelements<MAXELEMENTS) {
      lbuffer[lelements]=item;
      lelements++;
    }
  } else {
    if (relements<MAXELEMENTS) {
      rbuffer[relements]=item;
      relements++;
    }
  }
}

/*--- drawing section ---*/

void DrawPixel(int pg,int x,int y,int c) {
  if (c<0)
    c=WHITE;
  AddGE(pg,new GEpoint(x,y,c));
}

void DrawLine(int pg,int x1,int y1,int x2,int y2,int c) {
  if (c<0)
    c=WHITE;
  AddGE(pg,new GEline(x1,y1,x2,y2,c));
}

void DrawRectangle(int pg,int x1,int y1,int x2,int y2,int c=-1) {
  if (c<0)
    c=WHITE;
  AddGE(pg,new GErect(x1,y1,x2,y2,c));
}

static long int FILESIZE(char *fname) {
  struct stat status;

  stat(fname,&status);
  return status.st_size;
}

void RunServer() {
  clock_t t1,t2;
  word dt;
  int result;
  byte buffer[256];
  int *ibuffer;
  struct {
    long mtype;
    char *mtext;
  } mbuf;
/*  struct msqid_ds mstat; */

  ibuffer=(int*)buffer;
  side=RIGHT;
  SetPage(side);
  do {
    ClearCRT();
    delay(tcrt);
    if (side==LEFT) {
      ShutterRight();
      side=RIGHT;
    } else {
      ShutterLeft();
      side=LEFT;
    }
    delay(tlcd);
    SetPage(side);
    FlushPage();
    /*--- here is the right place to accept new input ---*/
    t1=clock();
    /*
      if (msgctl(channel,IPC_STAT,&mstat)!=-1) {
      if (mstat.msg_cbytes>0) {
      printf("current number of bytes in que: %d\n",mstat.msg_cbytes);
      printf("     number of messages in que: %d\n",mstat.msg_qnum);
      printf("    max number of bytes in que: %d\n",mstat.msg_qbytes);
      printf("            pid of last msgsnd: %d\n",mstat.msg_lspid);
      printf("              last receive pid: %d\n",mstat.msg_lrpid);
      }
      }
    */
    result=msgrcv(channel,(msgbuf*)(&mbuf),1024,0,IPC_NOWAIT);
    if (result>0) {
      ibuffer=(int*)(&(mbuf.mtext));
//      printf("command: %d\n",ibuffer[2]);
//      printf("%d %d %d %d %d\n",ibuffer[3],ibuffer[4],ibuffer[5],ibuffer[6],ibuffer[7]);
      switch (ibuffer[2]) {
        case LCDAPI_DRAW_LINE: DrawLine(ibuffer[3],ibuffer[4],ibuffer[5],ibuffer[6],ibuffer[7]);
	                       break;
        case LCDAPI_RESET:     ClearPage(LEFT);
	                       ClearPage(RIGHT);
                               break;
        case LCDAPI_FLUSH:     FlushBuffer();
	                       break;
      }
    }
    t2=clock();
    dt=tv-1000*(t2-t1)/CLOCKS_PER_SEC;
    if ((dt>0) && (dt<1000))
      delay(dt);
  } while (1);
}

int main(int argc,char **argv) {
  int w,h,c,t1,t2,t3;

	printf("LCD API daemon V 1.0\n(C) Nils Schmeisser, December 1996\n");
	printf("This software is restricted to the GNU General Public License Agreement\n");
  if (argc!=1) {
    sscanf(argv[1],"%d*%d*%d",&w,&h,&c);
    sscanf(argv[2],"%d",&t1);
    sscanf(argv[3],"%d",&t2);
    sscanf(argv[4],"%d",&t3);
    if (argc>5)
      sscanf(argv[5],"%d",&MAXELEMENTS);
    Initialize(w,h,c,t1,t2,t3);
  } else {
    printf("No arguments given, using default values: ");
    printf(" %c640*480*256%c 6 8 10\n",0x22,0x22);
    printf("usage: lcdd resolution tlcd tcrt tv\n");
    printf("       resolution %cwidth*height*colors%c\n",0x22,0x22);
    printf("       tlcd    shutter latency time [msec]\n");
    printf("       tcrt    crt/display latency time [msec]\n");
    printf("       tv      time to show image [msec]\n");
    Initialize(640,480,256,6,8,10);
  }

  DrawLine(LEFT ,320-100,240-100,320-10,240-10,WHITE);
  DrawLine(RIGHT,320- 80,240- 80,320- 8,240- 8,WHITE);

  DrawLine(LEFT ,320+100,240-100,320+10,240-10,WHITE);
  DrawLine(RIGHT,320+ 80,240- 80,320+ 8,240- 8,WHITE);

  DrawLine(LEFT ,320-100,240+100,320-10,240+10,WHITE);
  DrawLine(RIGHT,320- 80,240+ 80,320- 8,240+ 8,WHITE);

  DrawLine(LEFT ,320+100,240+100,320+10,240+10,WHITE);
  DrawLine(RIGHT,320+ 80,240+ 80,320+ 8,240+ 8,WHITE);

  DrawRectangle(LEFT ,320-10,240-10,320+10,240+10);
  DrawRectangle(RIGHT,320- 8,240- 8,320+ 8,240+ 8);

  DrawRectangle(LEFT ,320-100,240-100,320+100,240+100);
  DrawRectangle(RIGHT,320- 80,240- 80,320+ 80,240+ 80);

  FlushBuffer();
  SetPage(LEFT);
  FlushPage();
  SetPage(RIGHT);
  FlushPage();

  RunServer();
  return 0;
}






